/*
 * Copyright (C) 2003, 2004, 2005, 2006, 2007
 * Robert Lougher <rob@lougher.org.uk>.
 *
 * This file is part of JamVM.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "../../../jam.h"

#ifndef USE_FFI
#include <string.h>
#include "../../../sig.h"

#define RET_VOID    0
#define RET_DOUBLE  1
#define RET_LONG    2
#define RET_FLOAT   3
#define RET_DFLT    4

int nativeExtraArg(MethodBlock *mb) {
	//printf("!!!!!!!!!!nativeExtraArg (name=%s) (type=%s)!!!!!!!\n",mb->name,mb->type);
    int len = strlen(mb->type);
    if(mb->type[len-2] == ')')
        switch(mb->type[len-1]) {
            case 'V':
                return RET_VOID;
            case 'D':
                return RET_DOUBLE;
            case 'J':
                return RET_LONG;
            case 'F':
                return RET_FLOAT;
        }

    return RET_DFLT;
}

//calls the f function pointer and takes care of the return type
#define perform_f(label,...)													\
		switch(ret_type) {														\
        	case RET_VOID:														\
            	(*(void (*)())f)(__VA_ARGS__);									\
				goto label;														\
        	case RET_DOUBLE:													\
	        	*(double*)ostack = (*(double (*)())f)(__VA_ARGS__);				\
            	ostack += 2;													\
				goto label;														\
        	case RET_LONG:														\
	        	*(long long*)ostack = (*(long long (*)())f)(__VA_ARGS__);		\
            	ostack += 2;													\
				goto label;														\
        	case RET_FLOAT:														\
	        	*(float*)ostack = (*(float (*)())f)(__VA_ARGS__);				\
            	ostack++;														\
				goto label;														\
        	default:															\
	            *ostack++ = (*(u4 (*)())f)(__VA_ARGS__);						\
				goto label;														\
	    }

#define write_ops_to_stack(x)	    											\
	/*write x operands to the stack	*/											\
	int i;																		\
	for(i=0; i < x; i++){														\
		__asm__ __volatile__(													\
    	"ld.w r11,%0		\n\t"												\
    	"st.w --sp,r11		" 													\
    	:: "m"(*--opntr)														\
    	:"sp","r11");															\
	}										
	
u4 *callJNIMethod(void *env, Class *class, char *sig, int ret_type, u4 *ostack, unsigned char *f, int args) {
	u4 *opntr = ostack + args;
	int nrofopsonstack = 0;
	if(class && args > 3){
		nrofopsonstack = args - 3;
	}else if(class){
		nrofopsonstack = 0;
	}else if(args > 4){
		nrofopsonstack = args -4;
	}else{
		nrofopsonstack = 0;
	}
	
	if(args == 0){
		if(class){
			perform_f(done,env,class);
		}else{
			perform_f(done,env);
		}
	}else if(args == 1){
		if(class){
    		perform_f(done,env,class,*--opntr);
		}else{
    		perform_f(done,env,*--opntr);
		}
	}else if(args == 2){
		if(class){
    		perform_f(done,env,class,*--opntr,*--opntr);
		}else{
    		perform_f(done,env,*--opntr,*--opntr);
		}
	}else if(args == 3){
		if(class){
    		perform_f(done,env,class,*--opntr,*--opntr,*--opntr);
		}else{
    		perform_f(done,env,*--opntr,*--opntr,*--opntr);
		}
	}else{
		write_ops_to_stack(nrofopsonstack)
		if(class){
    		perform_f(empty_stack,env,class,*--opntr,*--opntr,*--opntr);
		}else{
    		perform_f(empty_stack,env,*--opntr,*--opntr,*--opntr,*--opntr);
		}
	}
	
	empty_stack:
	//Increment stack pointer to loose all operands we put there
	if(nrofopsonstack > 0) {
 		__asm__ __volatile__("add sp,%0" ::"r"((nrofopsonstack) * sizeof(u4)): "sp");
	}
		
	done:
	return ostack;
}

/*
    int i;
    int nrofopsonstack,nrofopsinregister;
	u4 *opntr = ostack + args;
	register int *r5 asm ("r5");
	*r5 = *f;
	
//	register unsigned char *r6 asm ("r6");
//	register int *r7 asm ("r7");
//	register int *r8 asm ("r8");
//	register int *r9 asm ("r9");
//	register int *r10 asm ("r10");
//	register int *r11 asm ("r11");
//	register int *r12 asm ("r12");
	
	if(class && args > 3){
		nrofopsonstack = args - 3;
		nrofopsinregister = 3;
	}else if(class){
		nrofopsonstack = 0;
		nrofopsinregister = args;
	}else if(args > 4){
		nrofopsonstack = args -4;
		nrofopsinregister = 4;
	}else{
		nrofopsonstack = 0;
		nrofopsinregister = args;
	}

	//write all operands except 3 or 4 to the stack	
	for(i=0; i < nrofopsonstack; i++){
		//printf("!!!!!writing arg %d to stack\n",i);
    	__asm__ __volatile__(
    	"ld.w r11,%0		\n\t"
    	"st.w --sp,r11		" 
    	:: "m"(*--opntr)
    	:"sp","r11");
	}

	
	//TODO: make sure none of the registers above have been changed
	switch(ret_type) {
        case RET_VOID:
        	goto do_void_function;
        case RET_DOUBLE:
        	goto do_double_function;
        case RET_LONG:
        	goto do_longlong_function;
        case RET_FLOAT:
        	goto do_float_function;
        default:
        	goto do_default_function;
    }

	do_void_function:
			WRITE_PARAMS(label_void)
			__asm__ __volatile__(
				"icall r5"
				::"m"(f));
			label_void:
            (*(void (*)())f)();
            //printf("!!!!!!!!!!!!! called void JNIMethod %s!!!!!!!!!!\n",sig);
            goto empty_stack;
            
	do_double_function:
			WRITE_PARAMS(label_double)
			label_double:
        	*(double*)ostack = (*(double (*)())f)();
            ostack += 2;
            //printf("!!!!!!!!!!!!! called double JNIMethod %s %l!!!!!!!!!!\n",sig,*ostack);
            goto empty_stack;

	do_longlong_function:
			WRITE_PARAMS(label_longlong)
			label_longlong:
        	*(long long*)ostack = (*(long long (*)())f)();
            ostack += 2;
            //printf("!!!!!!!!!!!!! called long JNIMethod %s %l!!!!!!!!!!\n",sig,*ostack);
            goto empty_stack;

	do_float_function:
			WRITE_PARAMS(label_float)
			label_float:
        	*(float*)ostack = (*(float (*)())f)();
            ostack++;
            //printf("!!!!!!!!!!!!! called float JNIMethod %s %f!!!!!!!!!!\n",sig,*ostack);
            goto empty_stack;

	do_default_function:
			WRITE_PARAMS(label_default)
			label_default:
            *ostack++ = (*(u4 (*)())f)();
            //printf("!!!!!!!!!!!!! called default JNIMethod %s!!!!!!!!!!\n",sig);
        	goto empty_stack;
    
    empty_stack:

	//Get rid of the items on the stack
	if(nrofopsonstack > 0) {
 		__asm__ __volatile__("add sp,%0" ::"r"(nrofopsonstack): "sp");
	}
    return ostack;*/


//    	r8 = (u4*)(*--opntr);
//    	r9 = (u4*)(*--opntr);
//    	r10 = (u4*)(*--opntr);
//    	r11 = (u4*)(*--opntr);
//    	r12 = env;



/*    
   	//write all operands to register
    if(!class){
    	switch(nrofopsinregister){
    	case 4:
   		__asm__ __volatile__(
   			"ld.w r8,%[r8]		\n\t"
   			"ld.w r9,%[r9]		\n\t"
   			"ld.w r10,%[r10]	\n\t"
   			"ld.w r11,%[r11]	\n\t"
   			"ld.w r12,%[env]"
   			:: [r8]"m"(*--opntr),[r9]"m"(*--opntr),[r10]"m"(*--opntr),[r11]"m"(*--opntr),[env]"m"(env)
   			:"r8","r9","r10","r11","r12"
   			);
   			goto select_function;
    	case 3:
   		__asm__ __volatile__(
   			"ld.w r9,%[r9]		\n\t"
   			"ld.w r10,%[r10]	\n\t"
   			"ld.w r11,%[r11]	\n\t"
   			"ld.w r12,%[env]"
   			:: [r9]"m"(*--opntr),[r10]"m"(*--opntr),[r11]"m"(*--opntr),[env]"m"(env)
   			:"r9","r10","r11","r12"
   			);
   			goto select_function;
    	case 2:
   		__asm__ __volatile__(
   			"ld.w r10,%[r10]	\n\t"
   			"ld.w r11,%[r11]	\n\t"
   			"ld.w r12,%[env]"
   			:: [r10]"m"(*--opntr),[r11]"m"(*--opntr),[env]"m"(env)
   			:"r10","r11","r12"
   			);
   			goto select_function;
    	case 1:
   		__asm__ __volatile__(
   			"ld.w r11,%[r11]	\n\t"
   			"ld.w r12,%[env]"
   			::[r11]"m"(*--opntr),[env]"m"(env)
   			:"r11","r12"
   			);
   			goto select_function;
   		case 0:
   		__asm__ __volatile__(
   			"ld.w r12,%[env]"
   			::[env]"m"(env)
   			:"r12"
   			);
   			goto select_function;
   		}
    }else{//class needs to be written to r11 
    	switch(nrofopsinregister){
    	case 3:
   		__asm__ __volatile__(
   			"ld.w r8,%[r8]		\n\t"
   			"ld.w r9,%[r9]		\n\t"
   			"ld.w r10,%[r10]	\n\t"
   			"ld.w r11,%[class]	\n\t"
   			"ld.w r12,%[env]"
   			:: [r8]"m"(*--opntr),[r9]"m"(*--opntr),[r10]"m"(*--opntr),[class]"m"(class),[env]"m"(env)
   			:"r8","r9","r10","r11","r12"
   			);
   			goto select_function;
    	case 2:
   		__asm__ __volatile__(
   			"ld.w r9,%[r9]		\n\t"
   			"ld.w r10,%[r10]	\n\t"
   			"ld.w r11,%[class]	\n\t"
   			"ld.w r12,%[env]"
   			:: [r9]"m"(*--opntr),[r10]"m"(*--opntr),[class]"m"(class),[env]"m"(env)
   			:"r9","r10","r11","r12"
   			);
   			goto select_function;
    	case 1:
   		__asm__ __volatile__(
   			"ld.w r10,%[r10]	\n\t"
   			"ld.w r11,%[class]	\n\t"
   			"ld.w r12,%[env]"
   			:: [r10]"m"(*--opntr),[class]"m"(class),[env]"m"(env)
   			:"r10","r11","r12"
   			);
   			goto select_function;
    	case 0:
   		__asm__ __volatile__(
   			"ld.w r11,%[class]	\n\t"
   			"ld.w r12,%[env]"
   			:: [class]"m"(class),[env]"m"(env)
   			:"r10","r11","r12"
   			);
   			goto select_function;
    	}
    }
	select_function: //This label is put here to make some cleaner assembler jumps above
	*/

//    int isstatic;
//    if(class){
//    	isstatic = 1; 
//    }else{
//    	isstatic = 0;
//    }
//    printf("!!!!!!!!!!!!! calling JNIMethod (sig:%s) (type:%d) (argcount:%d) (static:%d)!!!!!!!!!!\n",sig,ret_type,args,isstatic);

//Dit werkt!!!    
//    if(args == 3 && isstatic == 0 && ret_type == RET_DFLT){
//       printf("I am here..\n");
//       *ostack++ = (*(u4 (*)())f)(env,*--opntr,*--opntr,*--opntr);
//	   return ostack;
//    }
/*
    u4 *opntr = ostack + args;
    long long *result;
	if(class){
		//printf("I am static!!!!!!!\n");
		if(args == 0){
      		//*ostack = (*(u4 (*)())f)(env,class);
      		*result = (*(long long (*)())f)(env,class);
    	}else if(args == 1){
       		*result = (*(long long (*)())f)(env,class,*--opntr);
       	}else if(args == 2){
       		*result = (*(long long (*)())f)(env,class,*--opntr,*--opntr);
       	}else if(args == 3){
       		*result = (*(long long (*)())f)(env,class,*--opntr,*--opntr,*--opntr);
       	}else if(args == 4){
       		*result = (*(long long (*)())f)(env,class,*--opntr,*--opntr,*--opntr,*--opntr);
       	}else if(args == 5){
       		*result = (*(long long (*)())f)(env,class,*--opntr,*--opntr,*--opntr,*--opntr,*--opntr);
       	}
	}else{
      	//printf("I am not static!!!!!!!\n");
		if(args == 0){
       		*result = (*(long long (*)())f)(env);
       	}else if(args == 1){
       		*result = (*(long long (*)())f)(env,*--opntr);
       	}else if(args == 2){
       		*result = (*(long long (*)())f)(env,*--opntr,*--opntr);
       	}else if(args == 3){
       		*result = (*(long long (*)())f)(env,*--opntr,*--opntr,*--opntr);
       	}else if(args == 4){
       		*result = (*(long long (*)())f)(env,*--opntr,*--opntr,*--opntr,*--opntr);
       	}else if(args == 5){
       		*result = (*(long long (*)())f)(env,*--opntr,*--opntr,*--opntr,*--opntr,*--opntr);
       	}
	}
	
	switch(ret_type) {
        case RET_VOID:
			printf("-------->Performed void function %s\n",sig);
            break;

        case RET_DOUBLE:
        	*(double*)ostack = *(double(*))result;
        	ostack += 2;
			printf("-------->Performed double function %s,casting to %d\n",sig, ret_type);
        	//printf("returned a double %f\n",*ostack);
            break;

        case RET_LONG:
        	*(long long*)ostack = *(long long(*))result;
        	ostack += 2;
			printf("-------->Performed long function %s,casting to %d\n",sig, ret_type);
        	//printf("returned a long %l\n",*ostack);
            break;

        case RET_FLOAT:
        	*(float*)ostack = *(float(*))result;
        	ostack++;
			printf("-------->Performed float function %s,casting to %d\n",sig, ret_type);
        	//printf("returned a float %f\n",*ostack);
            break;

        default:
        	*ostack++ = *(u4(*))result;
            //ostack++;
			printf("-------->Performed default function %s,casting to %d\n",sig, ret_type);
        	//printf("returned a default %d\n",*ostack);
        	break;
    }
    return ostack;
    */




//    if(args == 3 && isstatic == 0 && ret_type == RET_DFLT){
//       printf("I am here!!!!!!!\n");
//       u4 *result = (*(u4 (*)())f)(env,*--opntr,*--opntr,*--opntr);
//       *ostack++ = result;
//       //*ostack++ = (*(u4 (*)())f)(env,*--opntr,*--opntr,*--opntr);
//	   return ostack;
//    }
//    if(args == 2 && isstatic == 0 && ret_type == RET_DFLT){
//       printf("I am here..\n");
//       *ostack++ = (*(u4 (*)())f)(env,*--opntr,*--opntr);
//	   return ostack;
//    }
//    if(args == 1 && isstatic == 0 && ret_type == RET_DFLT){
//       printf("I am here..\n");
//       *ostack++ = (*(u4 (*)())f)(env,*--opntr);
//	   return ostack;
//    }
//    
//    if(args == 6 && isstatic == 1 && ret_type == RET_DFLT){
//       printf("I am here for 7 args..\n");
//       *ostack++ = (*(u4 (*)())f)(env,class,*--opntr,*--opntr,*--opntr,*--opntr,*--opntr,*--opntr);
//	   return ostack;
//    }
//    if(args == 4 && isstatic == 1 && ret_type == RET_DFLT){
//       printf("I am here for 5 args..\n");
//       *ostack++ = (*(u4 (*)())f)(env,class,*--opntr,*--opntr,*--opntr,*--opntr);
//	   return ostack;
//    }
//    if(args == 3 && isstatic == 1 && ret_type == RET_DFLT){
//       printf("I am here..\n");
//       *ostack++ = (*(u4 (*)())f)(env,class,*--opntr,*--opntr,*--opntr);
//	   return ostack;
//    }
//    if(args == 2 && isstatic == 1 && ret_type == RET_DFLT){
//       printf("I am here..\n");
//      *ostack++ = (*(u4 (*)())f)(env,class,*--opntr,*--opntr);
//	   return ostack;
//   }
//    if(args == 1 && isstatic == 1 && ret_type == RET_DFLT){
//       printf("I am here..\n");
//       *ostack++ = (*(u4 (*)())f)(env,class,*--opntr);
//	   return ostack;
//    }
    
    //int i;

    //for(i = 0; i < args; i++){
    //	printf("!!!!!should write arg %d\n",i);
    	//__asm__ __volatile__("ld.w r11,%0" :: "m"(*--opntr):"r11");
        //__asm__ __volatile__("st.w --sp,r11" ::: "sp","r11");
    //}

//	int i=args;
//    if(class) {
//    	for(i=args;i>3;i--){
//	    	__asm__ __volatile__("ld.w r8,%0" :: "m"(*--opntr):"r8");
//    		__asm__ __volatile__("st.w --sp,r8" ::: "sp","r8");
//    	}
//    	if(args >= 3){
//    		__asm__ __volatile__("ld.w r8,%0" :: "m"(*--opntr):"r8");
//    	}
//    	if(args >= 2){
//    		__asm__ __volatile__("ld.w r9,%0" :: "m"(*--opntr):"r9");
//    	}
//    	if(args >= 1){
//    		__asm__ __volatile__("ld.w r10,%0" :: "m"(*--opntr):"r10");
//    	}
//    	__asm__ __volatile__("ld.w r11,%0" :: "m"(class):"r11");
//	   	__asm__ __volatile__("ld.w r12,%0" :: "m"(env):"r12");
//    }else{//static, don't add class variable
//	   	for(i=args;i>4;i--){
//	   		//r8 will be reused by the last register argument
//    		__asm__ __volatile__("ld.w r8,%0" :: "m"(*--opntr):"r8");
//    		__asm__ __volatile__("st.w --sp,r8" ::: "sp","r8");
//    	}
//    	if(args >= 4){
//    		__asm__ __volatile__("ld.w r8,%0" :: "m"(*--opntr):"r8");
//    	}
//    	if(args >= 3){
//    		__asm__ __volatile__("ld.w r9,%0" :: "m"(*--opntr):"r9");
//    	}
//    	if(args >= 2){
//    		__asm__ __volatile__("ld.w r10,%0" :: "m"(*--opntr):"r10");
//    	}
//    	if(args >= 1){
//    		__asm__ __volatile__("ld.w r11,%0" :: "m"(*--opntr):"r11");
//    	}
//    	__asm__ __volatile__("ld.w r12,%0" :: "m"(env):"r12");
//    }

//    printf("!!!!!pushing environment to r12\n");

    //printf("!!!!!!!!!!!!! switching ret_type %s %d ,nr of args: %d!!!!!!!!!!\n",sig, ret_type,args);
//    switch(ret_type) {
//        case RET_VOID:
//            (*(void (*)())f)();
//            //printf("!!!!!!!!!!!!! called void JNIMethod %s!!!!!!!!!!\n",sig);
//            break;
//
//        case RET_DOUBLE:
//        	*(double*)ostack = (*(double (*)())f)();
//            ostack += 2;
//            //printf("!!!!!!!!!!!!! called double JNIMethod %s %l!!!!!!!!!!\n",sig,*ostack);
//            break;
//
//        case RET_LONG:
//        	*(long long*)ostack = (*(long long (*)())f)();
//            ostack += 2;
//            //printf("!!!!!!!!!!!!! called long JNIMethod %s %l!!!!!!!!!!\n",sig,*ostack);
//            break;
//
//        case RET_FLOAT:
//        	*(float*)ostack = (*(float (*)())f)();
//            ostack++;
//            //printf("!!!!!!!!!!!!! called float JNIMethod %s %f!!!!!!!!!!\n",sig,*ostack);
//            break;
//
//        default:
//            *ostack++ = (*(u4 (*)())f)();
//            //printf("!!!!!!!!!!!!! called default JNIMethod %s!!!!!!!!!!\n",sig);
//        	break;
//    }
//
// if(class && args > 3) {
// 		int to_add = args-3;
// 		__asm__ __volatile__("add sp,%0" ::"r"(to_add): "sp");
// }else if(!class && args > 4){
// 		int to_add = args-4;
// 		__asm__ __volatile__("add sp,%0" ::"r"(to_add): "sp");
// }
//    //printf("!!!!!!!!!!!!! returning JNIMethod (sig:%s) (type:%d) (argcount:%d)!!!!!!!!!!\n",sig,ret_type,args);
//    //printf("!!!!!!!!!!!!! returning %s %d ,nr of args: %d!!!!!!!!!!\n",sig, ret_type,args);
//    //printf("!!!!!!!!!!!!! returning JNIMethod !!!!!!!!!!\n");
//    //__asm__ __volatile__ ("addl %0,%%esp" :: "r" ((args + 1) * sizeof(u4)) : "cc", "sp");
//    
//    return ostack;


//write all operands, class and env to register
#define WRITE_PARAMS(label)								\
    if(!class){											\
    	switch(nrofopsinregister){						\
    	case 4:											\
   		__asm__ __volatile__(							\
   			"ld.w r8,%[r8]		\n\t"					\
   			"ld.w r9,%[r9]		\n\t"					\
   			"ld.w r10,%[r10]	\n\t"					\
   			"ld.w r11,%[r11]	\n\t"					\
   			"ld.w r12,%[env]"							\
   			:: [r8]"m"(*--opntr),[r9]"m"(*--opntr),[r10]"m"(*--opntr),[r11]"m"(*--opntr),[env]"m"(env)						\
   			:"r8","r9","r10","r11","r12"				\
   			);											\
   			goto label;									\
    	case 3:											\
   		__asm__ __volatile__(							\
   			"ld.w r9,%[r9]		\n\t"					\
   			"ld.w r10,%[r10]	\n\t"					\
   			"ld.w r11,%[r11]	\n\t"					\
   			"ld.w r12,%[env]"							\
   			:: [r9]"m"(*--opntr),[r10]"m"(*--opntr),[r11]"m"(*--opntr),[env]"m"(env)						\
   			:"r9","r10","r11","r12"						\
   			);											\
   			goto label;									\
    	case 2:											\
   		__asm__ __volatile__(							\
   			"ld.w r10,%[r10]	\n\t"					\
   			"ld.w r11,%[r11]	\n\t"					\
   			"ld.w r12,%[env]"							\
   			:: [r10]"m"(*--opntr),[r11]"m"(*--opntr),[env]"m"(env)						\
   			:"r10","r11","r12"							\
   			);											\
   			goto label;									\
    	case 1:											\
   		__asm__ __volatile__(							\
   			"ld.w r11,%[r11]	\n\t"					\
   			"ld.w r12,%[env]"							\
   			::[r11]"m"(*--opntr),[env]"m"(env)			\
   			:"r11","r12"								\
   			);											\
   			goto label;									\
   		case 0:											\
   		__asm__ __volatile__(							\
   			"ld.w r12,%[env]"							\
   			::[env]"m"(env)								\
   			:"r12"										\
   			);											\
   			goto label;									\
   		}												\
    }else{												\
    	switch(nrofopsinregister){						\
    	case 3:											\
   		__asm__ __volatile__(							\
   			"ld.w r8,%[r8]		\n\t"					\
   			"ld.w r9,%[r9]		\n\t"					\
   			"ld.w r10,%[r10]	\n\t"					\
   			"ld.w r11,%[class]	\n\t"					\
   			"ld.w r12,%[env]"							\
   			:: [r8]"m"(*--opntr),[r9]"m"(*--opntr),[r10]"m"(*--opntr),[class]"m"(class),[env]"m"(env)						\
   			:"r8","r9","r10","r11","r12"				\
   			);											\
   			goto label;									\
    	case 2:											\
   		__asm__ __volatile__(							\
   			"ld.w r9,%[r9]		\n\t"					\
   			"ld.w r10,%[r10]	\n\t"					\
   			"ld.w r11,%[class]	\n\t"					\
   			"ld.w r12,%[env]"							\
   			:: [r9]"m"(*--opntr),[r10]"m"(*--opntr),[class]"m"(class),[env]"m"(env)						\
   			:"r9","r10","r11","r12"						\
   			);											\
   			goto label;									\
    	case 1:											\
   		__asm__ __volatile__(							\
   			"ld.w r10,%[r10]	\n\t"					\
   			"ld.w r11,%[class]	\n\t"					\
   			"ld.w r12,%[env]"							\
   			:: [r10]"m"(*--opntr),[class]"m"(class),[env]"m"(env)						\
   			:"r10","r11","r12"							\
   			);											\
   			goto label;									\
    	case 0:											\
   		__asm__ __volatile__(							\
   			"ld.w r11,%[class]	\n\t"					\
   			"ld.w r12,%[env]"							\
   			:: [class]"m"(class),[env]"m"(env)			\
   			:"r10","r11","r12"							\
   			);											\
   			goto label;									\
    	}												\
    }													\
	//This label is put here to make some cleaner assembler jumps above
#endif
